@@ -1,9 +1,10 @@
 """详细文档阅读地址: https://baidu.gitee.io/amis/zh-CN/components"""
-from typing import Union, List, Optional, Any
+import os
+from typing import Union, List, Optional, Any, Dict
 
 from pydantic import Field
 
-from .constants import LevelEnum, DisplayModeEnum, SizeEnum
+from .constants import LevelEnum, DisplayModeEnum, SizeEnum, TabsModeEnum
 from .types import API, Expression, AmisNode, SchemaNode, Template, BaseAmisModel, OptionsNode, Tpl
 from .utils import amis_templates
 
@@ -12,6 +13,8 @@ try:
 except ImportError:
     from typing_extensions import Literal
 
+BASE_DIR = os.path.dirname(os.path.abspath(__file__))
+
 
 class Html(AmisNode):
     """Html"""
@@ -53,6 +56,8 @@ class Badge(AmisNode):
 
 class Page(AmisNode):
     """页面"""
+    __default_template_path__: str = f'{BASE_DIR}/templates/page.html'
+
     type: str = "page"  # 指定为 Page 组件
     title: SchemaNode = None  # 页面标题
     subTitle: SchemaNode = None  # 页面副标题
@@ -77,11 +82,27 @@ class Page(AmisNode):
     stopAutoRefreshWhen: Expression = None  # 通过表达式来配置停止刷新的条件
     regions: List[str] = None
 
-    def amis_html(self, template_path: str = '', locale: str = ''):
+    def amis_html(
+            self,
+            template_path: str = '',
+            locale: str = 'zh_CN',
+            cdn: str = 'https://unpkg.com',
+            pkg: str = 'amis@1.10.1',
+            site_title: str = 'Amis',
+            site_icon: str = '',
+    ):
         """渲染html模板"""
-        return amis_templates('page.html', template_path).replace(
-            '[[AmisSchemaJson]]', self.amis_json()
-        ).replace('[[locale]]', locale)
+        template_path = template_path or self.__default_template_path__
+        return amis_templates(template_path).safe_substitute(
+            {
+                "AmisSchemaJson": self.amis_json(),
+                "locale": locale,
+                "cdn": cdn,
+                "pkg": pkg,
+                "site_title": site_title,
+                "site_icon": site_icon,
+            }
+        )
 
 
 class Divider(AmisNode):
@@ -143,16 +164,19 @@ class Tabs(AmisNode):
 
     class Item(AmisNode):
         title: str = None  # Tab 标题
-        icon: Icon = None  # Tab 的图标
+        icon: Union[str, Icon] = None  # Tab 的图标
         tab: SchemaNode = None  # 内容区
         hash: str = None  # 设置以后将跟 url 的 hash 对应
         reload: bool = None  # 设置以后内容每次都会重新渲染，对于 crud 的重新拉取很有用
         unmountOnExit: bool = None  # 每次退出都会销毁当前 tab 栏内容
         className: str = None  # "bg-white b-l b-r b-b wrapper-md"  # Tab 区域样式
+        iconPosition: str = None  # "left"  # Tab 的图标位置 left / right
+        closable: bool = None  # False  # 是否支持删除，优先级高于组件的 closable
+        disabled: bool = None  # False  # 是否禁用
 
     type: str = "tabs"  # 指定为 Tabs 渲染器
     className: str = None  # 外层 Dom 的类名
-    mode: str = None  # 展示模式，取值可以是 line、card、radio、vertical
+    mode: str = None  # 展示模式，取值可以是 line、card、radio、vertical、chrome、simple、strong、tiled、sidebar
     tabsClassName: str = None  # Tabs Dom 的类名
     tabs: List[Item] = None  # tabs 内容
     source: str = None  # tabs 关联数据，关联后可以重复生成选项卡
@@ -160,7 +184,31 @@ class Tabs(AmisNode):
     toolbarClassName: str = None  # tabs 中工具栏的类名
     mountOnEnter: bool = None  # False  # 只有在点中 tab 的时候才渲染
     unmountOnExit: bool = None  # False  # 切换 tab 的时候销毁
-    scrollable: bool = None  # False  # 是否导航支持内容溢出滚动，vertical和chrome模式下不支持该属性；chrome模式默认压缩标签
+    scrollable: bool = None  # False  # 是否导航支持内容溢出滚动，vertical和chrome模式下不支持该属性；chrome模式默认压缩标签（属性废弃）
+    tabsMode: TabsModeEnum = None  # 展示模式，取值可以是 line、card、radio、vertical、chrome、simple、strong、tiled、sidebar
+    addable: bool = None  # False  # 是否支持新增
+    addBtnText: str = None  # "增加"  # 新增按钮文案
+    closable: bool = None  # False  # 是否支持删除
+    draggable: bool = None  # False  # 是否支持拖拽
+    showTip: bool = None  # False  # 是否支持提示
+    showTipClassName: str = None  # "'' "  # 提示的类
+    editable: bool = None  # False  # 收否可编辑标签名
+    sidePosition: str = None  # "left"  # sidebar 模式下，标签栏位置 left / right
+
+
+class Portlet(Tabs):
+    """门户栏目"""
+
+    class Item(Tabs.Item):
+        toolbar: SchemaNode = None  # tabs 中的工具栏，随 tab 切换而变化
+
+    type: str = "portlet"  # 指定为 Portlet 渲染器
+    contentClassName: str = None  # Tabs content Dom 的类名
+    tabs: List[Item] = None  # tabs 内容
+    style: Union[str, dict] = None  # 自定义样式
+    description: Template = None  # 标题右侧信息
+    hideHeader: bool = None  # False  # 隐藏头部
+    divider: bool = None  # False  # 去掉分隔线
 
 
 class Horizontal(AmisNode):
@@ -172,7 +220,7 @@ class Horizontal(AmisNode):
 class Action(AmisNode):
     """行为按钮"""
     type: str = "button"  # 指定为 Page 渲染器。 button  action
-    actionType: str  # 【必填】这是 action 最核心的配置，来指定该 action 的作用类型，支持：ajax、link、url、drawer、dialog、confirm、cancel、prev、next、copy、close。
+    actionType: str = None  # 【必填】这是 action 最核心的配置，来指定该 action 的作用类型，支持：ajax、link、url、drawer、dialog、confirm、cancel、prev、next、copy、close。
     label: str = None  # 按钮文本。可用 ${xxx} 取值。
     level: LevelEnum = None  # 按钮样式，支持：link、primary、secondary、info、success、warning、danger、light、dark、default。
     size: str = None  # 按钮大小，支持：xs、sm、md、lg。
@@ -193,6 +241,7 @@ class Action(AmisNode):
         bool, str] = None  # 当action配置在dialog或drawer的actions中时，配置为true指定此次操作完后关闭当前dialog或drawer。当值为字符串，并且是祖先层弹框的名字的时候，会把祖先弹框关闭掉。
     required: List[str] = None  # 配置字符串数组，指定在form中进行操作之前，需要指定的字段名的表单项通过验证
     # primary:bool=None
+    onClick: str = None  # 自定义点击事件 通过字符串形式的 onClick 来定义点击事件，这个字符串会转成 JavaScript 函数
 
 
 class ActionType:
@@ -234,12 +283,12 @@ class ActionType:
 class PageSchema(AmisNode):
     """页面配置"""
     label: str = None  # 菜单名称。
-    icon: str = None  # 菜单图标，比如：fa fa-file.
+    icon: str = 'fa fa-flash'  # 菜单图标，比如：'fa fa-file'.
     url: str = None  # 页面路由路径，当路由命中该路径时，启用当前页面。当路径不是 / 打头时，会连接父级路径。
     # 比如：父级的路径为 folder，而此时配置 pageA, 那么当页面地址为 /folder/pageA 时才会命中此页面。
     # 当路径是 / 开头如： /crud/list 时，则不会拼接父级路径。
     # 另外还支持 /crud/view/:id 这类带参数的路由，页面中可以通过 ${params.id} 取到此值。
-    schema_: Page = Field(None, alias='schema')  # 页面的配置，具体配置请前往 Page 页面说明
+    schema_: Union[Page, "Iframe"] = Field(None, alias='schema')  # 页面的配置，具体配置请前往 Page 页面说明
     schemaApi: API = None  # 如果想通过接口拉取，请配置。返回路径为 json>data。schema 和 schemaApi 只能二选一。
     link: str = None  # 如果想配置个外部链接菜单，只需要配置 link 即可。
     redirect: str = None  # 跳转，当命中当前页面时，跳转到目标页面。
@@ -250,9 +299,31 @@ class PageSchema(AmisNode):
     children: List["PageSchema"] = None  # 子菜单
     sort: int = None  # 排序
 
+    def as_tabs_item(self, tabs_extra: Dict[str, Any] = None, item_extra: Dict[str, Any] = None):
+        if self.children:
+            tab = Tabs(
+                tabs=[item.as_tabs_item(tabs_extra, item_extra) for item in self.children]
+            ).update_from_dict(tabs_extra or {})
+        elif self.schema_:
+            tab = self.schema_
+            if isinstance(tab, Iframe):
+                tab.height = 1080
+        elif self.schemaApi:
+            tab = Service(schemaApi=self.schemaApi)
+        elif self.link:
+            tab = Page(body=Link(href=self.link, body=self.label, blank=True))
+        else:
+            tab = None
+        return Tabs.Item(
+            title=self.label,
+            icon=self.icon,
+            tab=tab,
+        ).update_from_dict(item_extra or {})
 
-class App(AmisNode):
+
+class App(Page):
     """多页应用"""
+    __default_template_path__: str = f'{BASE_DIR}/templates/app.html'
     type: str = "app"
     api: API = None  # 页面配置接口，如果你想远程拉取页面配置请配置。返回配置路径 json>data>pages，具体格式请参考 pages 属性。
     brandName: str = None  # 应用名称
@@ -266,12 +337,6 @@ class App(AmisNode):
 
     # 通常为数组，数组第一层为分组，一般只需要配置 label 集合，如果你不想分组，直接不配置，真正的页面请在第二层开始配置，即第一层的 children 中。
 
-    def amis_html(self, template_path: str = '', locale: str = ''):
-        """渲染html模板"""
-        return amis_templates('app.html', template_path).replace(
-            '[[AmisSchemaJson]]', self.amis_json()
-        ).replace('[[locale]]', locale)
-
 
 class ButtonGroup(AmisNode):
     """按钮组"""
@@ -414,6 +479,7 @@ class FormItem(AmisNode):
     requiredOn: Expression = None  # 过表达式来配置当前表单项是否为必填。
     validations: Union[Validation, Expression] = None  # 表单项值格式验证，支持设置多个，多个规则用英文逗号隔开。
     validateApi: Expression = None  # 表单校验接口
+    copyable: Union[bool, dict] = None  # 是否可复制  boolean 或 {icon: string, content:string}
 
 
 class Form(AmisNode):
@@ -719,7 +785,6 @@ class InputTable(FormItem):
     addable: bool = None  # False  # 是否可增加一行
     editable: bool = None  # False  # 是否可编辑
     removable: bool = None  # False  # 是否可删除
-    copyable: bool = None  # False  # 是否可复制
     showAddBtn: bool = None  # True  # 是否显示添加按钮
     addApi: API = None  # 新增时提交的 API
     updateApi: API = None  # 修改时提交的 API
@@ -1257,6 +1322,7 @@ class CRUD(AmisNode):
     alwaysShowPagination: bool = None  # 是否总是显示分页
     affixHeader: bool = None  # True  # 是否固定表头(table 下)
     autoGenerateFilter: bool = None  # 是否开启查询区域，开启后会根据列元素的 searchable 属性值，自动生成查询条件表单
+    itemAction: Action = None  # 实现点击某一行后进行自定义操作，支持 action 里的所有配置，比如弹框、刷新其它组件等。
 
 
 class TableColumn(AmisNode):
@@ -1267,7 +1333,7 @@ class TableColumn(AmisNode):
     name: str = None  # 通过名称关联数据
     tpl: Template = None  # 模板
     fixed: str = None  # 是否固定当前列 left|right|none
-    popOver: bool = None  # 弹出框
+    popOver: Union[bool, dict] = None  # 弹出框
     quickEdit: Union[bool, dict] = None  # 快速编辑
     copyable: Union[bool, dict] = None  # 是否可复制  boolean 或 {icon: string, content:string}
     sortable: bool = None  # False  # 是否可排序
@@ -1314,7 +1380,7 @@ class Table(AmisNode):
     itemActions: List[Action] = None  # 悬浮行操作按钮组
     itemCheckableOn: Expression = None  # 配置当前行是否可勾选的条件，要用 表达式
     itemDraggableOn: Expression = None  # 配置当前行是否可拖拽的条件，要用 表达式
-    checkOnItemClick: bool = False  # 点击数据行是否可以勾选当前行
+    checkOnItemClick: bool = None  # False  # 点击数据行是否可以勾选当前行
     rowClassName: str = None  # 给行添加 CSS 类名
     rowClassNameExpr: Template = None  # 通过模板给行添加 CSS 类名
     prefixRow: list = None  # 顶部总结行
@@ -1499,8 +1565,8 @@ class Iframe(AmisNode):
     frameBorder: list = None  # frameBorder
     style: dict = None  # 样式对象
     src: str = None  # iframe 地址
-    height: Union[int, str] = None  # iframe 高度
-    width: Union[int, str] = None  # iframe 宽度
+    height: Union[int, str] = None  # "100%"#  iframe 高度
+    width: Union[int, str] = None  # "100%" # iframe 宽度
 
 
 class Spinner(AmisNode):
